(load "Macintosh HD:Users:costanza:lisp:develop:closer:contextl:test:demo3")
(load "Macintosh HD:Users:costanza:lisp:develop:closer:contextl:test:spx")
(load "Macintosh HD:Users:costanza:lisp:develop:closer:contextl:test:grouped-layers")
(load "Macintosh HD:Users:costanza:lisp:develop:closer:contextl:test:special-slots-1")
(load "Macintosh HD:Users:costanza:lisp:develop:closer:contextl:test:special-slots-2")
(load "Macintosh HD:Users:costanza:lisp:develop:closer:contextl:test:special-slots-3")
(load "Macintosh HD:Users:costanza:lisp:develop:closer:contextl:test:layer-gc")
(load "Macintosh HD:Users:costanza:lisp:develop:closer:contextl:test:dynenv")
(load "Macintosh HD:Users:costanza:lisp:develop:closer:contextl:test:partial-initargs")

;; figure-editor and figure-editor-2
