(load "/Users/costanza/lisp/closer/contextl/test/demo3")
(load "/Users/costanza/lisp/closer/contextl/test/spx")
(load "/Users/costanza/lisp/closer/contextl/test/grouped-layers")
(load "/Users/costanza/lisp/closer/contextl/test/special-slots-1")
(load "/Users/costanza/lisp/closer/contextl/test/special-slots-2")
(load "/Users/costanza/lisp/closer/contextl/test/special-slots-3")
(load "/Users/costanza/lisp/closer/contextl/test/layer-gc")
(load "/Users/costanza/lisp/closer/contextl/test/dynenv")

;; figure-editor and figure-editor-2
