(load "demo3.lisp")
(load "spx.lisp")
(load "grouped-layers.lisp")
(load "fe.lisp")
(load "fe2.lisp")
(load "special-slots-1.lisp")
(load "special-slots-2.lisp")
(load "special-slots-3.lisp")
(load "layer-gc.lisp")
(load "dynenv.lisp")
(load "partial-initargs.lisp")
